/*
 * shipout.c
 * output routines for fm
 * Greg Lee, 7/01
 */
#include "fm.h"

#ifdef HAVE_LIBNCURSES
#include <curses.h>
#include <term.h>
#else
#ifdef HAVE_LIBTERMCAP
#include <termcap.h>
#endif
#endif

#ifdef USESIGNALS
#include <signal.h>
#endif


#define BUFFERSIZE (96*1024)


void    terminit();
static void get_screensize(int *linep, int *colp);
static void open_getchr();

static void	termattr(int i);
static void	colorattr(int i);
#ifndef STANDALONETREE
static
#endif
void	newattr(unsigned char attr);
#ifndef STANDALONETREE
static void	do_space(int spaces);
static void	shipout(int purge);
static int	getchr();
static void	raw_mode(int on);
#endif

#ifdef STANDALONETREE

int
	baselineskip = 1,
        hsize  = -1,
        hsize_sp  = -1,
	cols = 80,
	cols_sp,
        tabsize = 8,
        tabid = -1,
	outdev = 0,
	hoffset = 0;
#endif

int in_is_tty = 0;
int out_is_tty = 0;
static int tty;

static int page_length = 25;
#ifndef STANDALONETREE
static int text_height = 22;
static int top_margin = 1;
static int bot_margin = 1;
static int curr_page = 1;
int pageno = 1;
char pageno_str[50];
void fix_pageno();
int romanize(int num, char *s);
int nopagenumbers = FALSE;
int pageattop = TRUE;
int pagestretch = 100;
static int curr_tb = 0;

#define VGLUE -1
#define VFILL -2
#define ENDBOX -100

#define HBOX_ATOM -1
#ifdef DONOTIFY
#define NOTIFY_ATOM -2
#endif
#define DEAD_ATOM -3
#define FOOTNOTE_ATOM -4

#define PURGE 1
#define NOPURGE 0

#define MAXHBOX 1000
#define MAXPAGE 256
#define MAXINSERT 256

/* reserve 0, 1 for headline and footline */
#define LOWEST_HBOX 2

struct hboxtype *hbox[MAXINSERT];
static int curr_hbox = LOWEST_HBOX;
static int first_hbox = LOWEST_HBOX;
static int next_ins_hboxes = 1;

static struct {
	int atom, which, count, first, maxlen,
	   total_lines, minspaces, peg;
} insertion[MAXINSERT];
static int curr_ins = 0;
static int pending_insertions = 0;

static unsigned char *buf = NULL, *bufp = NULL;
static unsigned char all_roman[MAXLINE];

static int buffered_hboxes = 0;

static int page_line[MAXPAGE];
static int last_page_line[MAXPAGE];
static int page_ins[MAXPAGE];
static int last_page_ins[MAXPAGE];
static int last_page_flag = FALSE;
static int ps_line[MAXPAGE];

int penalty_line[MAXPAGE];

static void NL() {
	if (psflag) return;
	if (out_is_tty && !fflag) putchar('\r');
	putchar('\n');
}
/* there has to be a better way ... */
int romanize(int num, char *s) {
	int j, i = 0;
	if (num < 0 && num > -25000) {
		j = -num;
		while (j >= 1000) {
			s[i++] = 'm';
			j -= 1000;
		}
		if (j >= 999) {
			s[i++] = 'i';
			s[i++] = 'm';
			j -= 999;
		}
		if (j >= 995) {
			s[i++] = 'v';
			s[i++] = 'm';
			j -= 995;
		}
		if (j >= 990) {
			s[i++] = 'x';
			s[i++] = 'm';
			j -= 995;
		}
		if (j >= 950) {
			s[i++] = 'l';
			s[i++] = 'm';
			j -= 995;
		}
		if (j >= 900) {
			s[i++] = 'c';
			s[i++] = 'm';
			j -= 900;
		}
		if (j >= 500) {
			s[i++] = 'd';
			j -= 500;
		}
		if (j >= 499) {
			s[i++] = 'i';
			s[i++] = 'd';
			j -= 499;
		}
		if (j >= 495) {
			s[i++] = 'v';
			s[i++] = 'd';
			j -= 495;
		}
		if (j >= 490) {
			s[i++] = 'x';
			s[i++] = 'd';
			j -= 490;
		}
		if (j >= 450) {
			s[i++] = 'l';
			s[i++] = 'd';
			j -= 450;
		}
		if (j >= 400) {
			s[i++] = 'c';
			s[i++] = 'd';
			j -= 400;
		}
		while (j >= 100) {
			s[i++] = 'c';
			j -= 100;
		}
		if (j >= 99) {
			s[i++] = 'i';
			s[i++] = 'c';
			j -= 99;
		}
		if (j >= 95) {
			s[i++] = 'v';
			s[i++] = 'c';
			j -= 95;
		}
		if (j >= 90) {
			s[i++] = 'x';
			s[i++] = 'c';
			j -= 90;
		}
		if (j >= 50) {
			s[i++] = 'l';
			j -= 50;
		}
		if (j >= 49) {
			s[i++] = 'i';
			s[i++] = 'l';
			j -= 49;
		}
		if (j >= 40) {
			s[i++] = 'x';
			s[i++] = 'l';
			j -= 40;
		}
		while (j >= 10) {
			s[i++] = 'x';
			j -= 10;
		}
		if (j >= 9) {
			s[i++] = 'i';
			s[i++] = 'x';
			j -= 9;
		}
		if (j >= 5) {
			s[i++] = 'v';
			j -= 5;
		}
		if (j >= 4) {
			s[i++] = 'i';
			s[i++] = 'v';
			j -= 4;
		}
		while (j >= 1) {
			s[i++] = 'i';
			j -= 1;
		}
		s[i] = 0;
	}
	else sprintf(s, "%d", num);
	return strlen(s);
}
void fix_pageno() {
	romanize(pageno, pageno_str);
}

static int starting_numbering = 0;
void next_pageno() {
	curr_page++;
	if (!nopagenumbers) {
		if (pageno >= 0) pageno++;
		else pageno--;
		fix_pageno();
	}
	if (nopagenumbers < 0) {
	    starting_numbering = nopagenumbers - 2;
	    nopagenumbers++;
	}
	if (starting_numbering < 0) {
	    starting_numbering++;
	}
}
void prev_pageno() {
	curr_page--;
	if (!nopagenumbers) {
		if (pageno > 0) pageno--;
		else if (pageno < -1) pageno++;
		fix_pageno();
		if (starting_numbering) nopagenumbers--;
	}
}
static int next_hbox() {
	curr_hbox++;
	if (curr_hbox == MAXHBOX) curr_hbox = LOWEST_HBOX;
	return curr_hbox;
}

static int prev_hbox(int i, int j) {
	int d = j - i;
	if (d >= 0 && d < MAXHBOX/2) return TRUE;
	if (d >= 0) return FALSE;
	if (-d < MAXHBOX/2) return FALSE;
	return TRUE;
}

static unsigned char *keep_this(unsigned char *src, int len) {
	int used = bufp - buf;
	int room = BUFFERSIZE - used;
	unsigned char *ret;

	if (used + len >= room) bufp = buf;
	ret = bufp;
	memcpy(bufp, src, len);
	bufp += len;
	return ret;
}
#endif

static char alt = 0;
static int curr_col = 0;
static void schar(int ch) {
	if (curr_col < cols || !out_is_tty) {
	    if (alt) putchar(ch);
	    else if (ch < ' ' || (ch > 126 && ch < 0xa0)) {
		if (ch > 126) {
			if (uflag) {
				if (ch > 127 + ' ') colorattr(c_super);
				else colorattr(c_sub);
			}
			ch -= 128;
		}
		else if (uflag) colorattr(c_prompt);
		if (ch < ' ') ch += 64;
		putchar(ch);
		if (uflag) colorattr(c_roman);
	    }
	    else putchar(ch);
	}
	curr_col++;
}
static void uchar(unsigned char *p, unsigned char *q, int len, int litflag) {
	int ch = *p;
	unsigned char a = *q;
	if (curr_col < cols || !out_is_tty) {
	    if (/* alt ||*/ ch == ' ' && !litflag && len > 0) putchar(ch);
	    else if (kflag && (len == -1 || ch >= 0xa0 || is_ipa(a) ||
		 a == MATH || a > 128 || a == BOX || a == ZD || a == SUPER || a == SUB)) {
			int diacritic = -1;
			unsigned char da = 0;
			if (len > 1 && a != MATH) {
				diacritic = *(p + 1);
				da = *(q + 1);
			}
			if (diacritic == LIT_ESC) {
				if (len > 2) {
					diacritic = *(p + 2);
					da = *(q + 2);
				}
				else diacritic = -1;
			}
			else if (is_ctl(diacritic)) diacritic = -1;

			if (len == -1) {
				a = da = 0;
				diacritic = *(p + 1);
			}

			if (!uni_out(ch, a, diacritic, da)) curr_col--;
	    }
	    else if (ch < ' ' || (ch > 126 && ch < 0xa0)) {
		if (ch > 126) {
			if (uflag) {
				if (ch > 127 + ' ') colorattr(c_super);
				else colorattr(c_sub);
			}
			ch -= 128;
		}
		else if (uflag) colorattr(c_prompt);
		if (ch < ' ') ch += 64;
		putchar(ch);
		if (uflag) colorattr(c_roman);
	    }
	    else putchar(ch);
	}
	curr_col++;
}

#ifdef STANDALONETREE
void pcnl() {
	if (uflag) newattr(ROMAN);
	putchar('\n');
	curr_col = 0;
}
void pch(unsigned char c, unsigned char attr) {
	unsigned char p[2], q[2];
	p[0] = c;
	q[0] = attr;
	p[1] = q[1] = '\0';
	uchar(p, q, 1, FALSE);
}
void pspace() {
	if (uflag) newattr(ROMAN);
	schar(' ');
}
void pstr(unsigned char *ostr, unsigned char *astr, int len) {

	int litflag, ch;
	unsigned char   *p;
	unsigned char   *q;
	unsigned char    a;

	for (p = ostr, q = astr; len; len--, p++, q++) {
		ch = *p;
		if (ch == '\t') { ch = ' '; a = ROMAN; }
		else a = *q;
		if (ch == LIT_ESC) {
			if (len <= 1) break;
			p++; q++; len--;
			a = *q;
			litflag = TRUE;
		}
		else litflag = FALSE;
		if (uflag) newattr(a);
		if (ch == UNI_ESC && !litflag) {
			if (len < 2) break;
			p++; q++; len--;
			uchar(p, q, -1, TRUE);
			p++; q++; len--;
		}
		else uchar(p, q, len, litflag);
	}
}
#endif

#ifndef STANDALONETREE
static void fakeputline (unsigned char *ostr, unsigned char *astr, int len) {
	int litflag;
	unsigned char   *p;
	unsigned char   *q;
	unsigned char    a, last = ROMAN;

	curr_col = offset;
	dotab(-1, curr_tb);
	for (p = ostr, q = astr; len; len--, p++, q++) {
		if (*p == '\t') a = ROMAN;
		else a = *q;
		if (*p == LIT_ESC) {
			if (len <= 1) break;
			p++; q++; len--;
			a = *q;
			litflag = TRUE;
		}
		else litflag = FALSE;
		if (a >= 128) a -= 128;
		if (!aflag) switch (a) {
			case BOX:
			case ROMAN: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: break;
				    case  BOLD: schar('*'); break;
				    case  MATH: schar('\''); break;
				    case TTYPE: schar('"'); break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: schar('/'); break;
				    case SUB:
				    case IPASUB:
				    case IPA:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
			case BOLD: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('*'); break;
				    case  BOLD: break;
				    case  MATH: schar('\''); break;
				    case TTYPE: schar('"');  break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: schar('/'); break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
	    		case  MATH:
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('`');  break;
				    case  BOLD: schar('*');  break;
	    			    case  MATH: break;
				    case TTYPE: schar('"');  break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: schar('/'); break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
			case TTYPE: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('"'); break;
				    case  BOLD: schar('*'); break;
				    case  MATH: schar('\''); break;
				    case TTYPE: break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: schar('/'); break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
			case ITALIC: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('_'); break;
				    case  BOLD: schar('*'); break;
				    case  MATH: schar('\''); break;
				    case TTYPE: schar('"');  break;
				    case ITALIC: break;
				    case SLANTY: schar('/'); break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
			case SLANTY: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('/'); break;
				    case  BOLD: schar('*'); break;
				    case  MATH: schar('\''); break;
				    case TTYPE: schar('"');  break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
		        case IPA:
			case SUB:
			case IPASUB:
			case BITALIC: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('['); break;
				    case  BOLD: schar('*'); break;
				    case  MATH: schar('\''); break;
				    case TTYPE: schar('"');  break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: schar('/'); break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: schar('}'); break;
				}
				break;
			case SUPER: 
			case IPASUPER: 
			case MUSIX:
			case SANS: 
				switch (last) {
				    case ZD:
				    case BOX:
				    case ROMAN: schar('{'); break;
				    case  BOLD: schar('*'); break;
				    case  MATH: schar('\''); break;
				    case TTYPE: schar('"');  break;
				    case ITALIC: schar('_'); break;
				    case SLANTY: schar('/'); break;
				    case IPA:
				    case SUB:
				    case IPASUB:
				    case BITALIC: schar(']'); break;
				    case SUPER:
				    case IPASUPER:
				    case MUSIX:
				    case SANS: break;
				}
				break;
		}
		if (*p == '\t' && !litflag) {
			curr_col = offset + dotab(curr_col - offset, curr_tb);
			last = ROMAN;
			continue;
		}
		if (*p == UNI_ESC && !litflag) {
				if (len < 2) break;
				p++; q++; len--;
				uchar(p, q, -1, FALSE);
				p++; q++; len--;
				continue;
		}
		uchar(p, q, len, litflag);
		last = a;
	}
	if (!aflag) switch (last) {
	    case ZD:
	    case BOX:
	    case ROMAN: break;
	    case  BOLD: schar('*');  break;
	    case  MATH: schar('\''); break;
	    case TTYPE: schar('"');  break;
	    case ITALIC: schar('_'); break;
	    case SLANTY: schar('/'); break;
	    case IPA:
	    case SUB:
	    case IPASUB:
	    case BITALIC: schar(']'); break;
	    case SUPER:
	    case IPASUPER:
	    case MUSIX:
	    case SANS: schar('}'); break;
	}

}				/* fakeputline */


/* display characters of line with overstriking or with special
 * terminal attributes */
static void sputline (unsigned char *ostr, unsigned char *astr, int len) {
	int litflag;
	unsigned char   *p;
	unsigned char   *q;
	unsigned char    a;

	dotab(-1, curr_tb);

	for (p = ostr, q = astr; len; len--, p++, q++) {
		if (*p == '\t') a = ROMAN;
		else a = *q;
		if (*p == LIT_ESC) {
			if (len <= 1) break;
			p++; q++; len--;
			a = *q;
			litflag = TRUE;
		}
		else litflag = FALSE;
		if (uflag) {
			newattr(a);
			if (*p == '\t' && !litflag)
				curr_col = offset + dotab(curr_col - offset, curr_tb);
			else if (*p == UNI_ESC && !litflag) {
				if (len < 2) break;
				p++; q++; len--;
				uchar(p, q, -1, FALSE);
				p++; q++; len--;
			}
			else uchar(p, q, len, litflag);
		} else {
		    if (a >= 128) a -= 128;
		    if (*p == '\t' && !litflag)
				curr_col = offset + dotab(curr_col - offset, curr_tb);
		    else if (*p == UNI_ESC && !litflag) {
				if (len < 2) break;
				p++; q++; len--;
				uchar(p, q, -1, FALSE);
				p++; q++; len--;
		    }
		    else {
		      switch (a) {
	    		case ZD:
	    		case BOX:
			case ROMAN: 
			case MUSIX:
			case SANS: 
				if (*p > 127) uchar(p, q, len, FALSE);
				else schar(*p);
				break;
			case MATH:
			case BOLD: 
			case BITALIC: 
			case TTYPE: 
				if (*p == ' ') schar(*p);
				else if (*p > 127 && kflag)
					uchar(p, q, len, FALSE);
				else {
				    if (curr_col<cols) printf("%c%c%c", *p, 8, *p);
				    curr_col++;
				}
				break;
			case ITALIC: 
			case SLANTY: 
				if (*p == ' ') schar(*p);
				else if (*p > 127 && kflag)
					uchar(p, q, len, FALSE);
				else {
				    if (curr_col<cols) printf("_%c%c", 8, *p);
				    curr_col++;
				}
				break;
			case IPA:
				if (*p == ' ') schar(*p);
				else if (kflag)
					uchar(p, q, len, FALSE);
				else {
				    if (curr_col<cols) printf("_%c%c", 8, *p);
				    curr_col++;
				}
				break;
		      }
		    }
		}
	}
	if (uflag) newattr(ROMAN);

}				/* sputline */

/* display attributes of line  */
static void reputline(int wh, int i) {
	unsigned char   *p;
	unsigned char   *q;
	int len = hbox[wh][i].len;

	for (p = hbox[wh][i].ch, q = hbox[wh][i].at; len; len--, p++, q++) {
		if (*p == LIT_ESC) {
			if (len <= 1) break;
			p++; q++; len--;
		}
		switch (*q) {
	    		case ZD:
	    		case BOX:
			case ROMAN: 
			case MUSIX:
			case SANS: 
				schar(' ');
				break;
			case BOLD: 
			case BITALIC: 
				schar('=');
				break;
			case MATH:
			case TTYPE: 
				schar('-');
				break;
			case ITALIC: 
			case IPA: 
				schar('^');
				break;
			case SLANTY: 
				schar('/');
				break;
			default:
				schar(' ');
				break;
		}
	}
}				/* reputline */

void windup() {
	if (insertion[curr_ins].count && !pending_insertions)
		pending_insertions++;
	shipout(PURGE);
	if (out_is_tty) {
		raw_mode(FALSE);
		if (kflag && !alwayskflag) printf("%c%%@", 27);
	}
	if (vflag) fprintf(stderr,"\n");
}

static void outline(int wh, int i)
{
	int len = hbox[wh][i].len, j;
	curr_tb = hbox[wh][i].tb;

	if (oflag || uflag)
			sputline(hbox[wh][i].ch, hbox[wh][i].at, len);
		else {
			if (!hbox[wh][i].us)
				 fakeputline(hbox[wh][i].ch, hbox[wh][i].at, len);
			else {
				char *s = hbox[wh][i].ch;
				for (j = 0; j < len; j++) schar(s[j]);
			}
		}
}

static int fn_lines = 0;

/* #define VJDEBUG */
static void vjustify(int under_skip) {
	int pline, i, j, vfillcnt = 0, vpads = 0;
	int end_line, dojustify = TRUE;

	if (fflag && !psflag) return;
	if (baselineskip > 1 && !psflag) return;

	end_line = text_height - fn_lines - 1;

	/* a vfill after last text stops justification */
	for (pline = end_line ; pline > 0; pline--) {
		if (page_line[pline] >= 0) break;
		if (page_line[pline] == VFILL) break;
	}
	if (pline == 0 && !psflag) dojustify = FALSE;
	if (page_line[pline] == VFILL) dojustify = FALSE;

	pline = 0;
	/* leave enough glue at bottom to soak up under_skip */
	if (psflag && under_skip > 0) j = 1 + under_skip / ps_baselineskip;
	else j = 0;

	if (dojustify) while (page_line[end_line] == VGLUE) {
		if (psflag && j > 0) {
			for (i = end_line-j; i > 0 && i < end_line; i++)
				if (page_line[i] >= 0) j = 0;
			if (!j) break;
		}
		for ( ; pline < end_line; pline++) {
			if (page_line[pline] == VFILL) break;
		}
		if (pline >= end_line) {
			if (!vfillcnt) break;
			pline = 0;
			continue;
		}
		vfillcnt++;
		for (i = end_line; i > pline; i--) {
			page_line[i] = page_line[i-1];
			page_ins[i] = page_ins[i-1];
			penalty_line[i] = penalty_line[i-1];
			/* ps_line[i] = ps_line[i-1] + ps_baselineskip; */
			ps_line[i] = ps_line[i-1];
		}
		pline++;
		page_line[pline] = VGLUE;
		pline++;
	}

	if (psflag) {
	    int jspace = 0, glue_count = 0, last_line, line_glue;
	    int accum_skip = 0;

	    for (pline = end_line; pline && page_line[pline] == VGLUE; pline--, jspace++) ;
#ifdef VJDEBUG
fprintf(stderr,"vjust: end_line=%d jspace=%d text_height=%d fn_lines=%d dojustify=%d\n", end_line, jspace,
		text_height, fn_lines, dojustify);
#endif
	    if (dojustify) {
	        jspace = jspace * ps_baselineskip10 / 10;
	        jspace -= under_skip;
	    }
	    else jspace = 0;

	    if (jspace) {
		last_line = pline;
		for (pline = 0; pline < last_line; pline++)
			if (page_line[pline] < 0 && penalty_line[pline] <= 0) glue_count++;
		if (glue_count) {
		    line_glue = jspace / glue_count;
		    jspace -= line_glue * glue_count;
		    for (pline = 0; pline < last_line; pline++) {
			if (page_line[pline] < 0 && penalty_line[pline] <= 0) {
				ps_line[pline] += line_glue;
				if (jspace > 0) {
					ps_line[pline] += 1;
					jspace--;
				}
				else if (jspace < 0) {
					ps_line[pline] -= 1;
					jspace++;
				}
#ifdef VJDEBUG
fprintf(stderr,"\tline %d gets glue %d, last_line=%d glue_count=%d jspace=%d\n",
		pline, line_glue, last_line, glue_count, jspace);
#endif
			}
		    }
		}
		if (jspace > 0 && under_skip) {
		    line_glue = jspace / under_skip;
		    jspace -= line_glue * under_skip;
		    for (pline = 0; pline < last_line; pline++) {
			    if (ps_line[pline]) {
				ps_line[pline] += line_glue;
				if (jspace > 0) {
					ps_line[pline] += 1;
					jspace--;
				}
			    }
		    }
		}
		if (jspace > 0) {
		    line_glue = jspace / last_line;
		    jspace -= line_glue * last_line;
#ifdef VJDEBUG
fprintf(stderr,"\teach line gets glue %d, last_line=%d glue_count=%d jspace=%d\n",
		line_glue, last_line, glue_count, jspace);
#endif
		    for (pline = 0; pline < last_line; pline++) {
				ps_line[pline] += line_glue;
				if (jspace > 0) {
					ps_line[pline] += 1;
					jspace--;
				}
		    }
		}
	    }

	    for (pline = 0; pline < text_height; pline++) {
		line_glue = ps_line[pline];
		ps_line[pline] = pline * ps_baselineskip10 / 10 + accum_skip;
		if (pline == end_line) accum_skip = 0;
		accum_skip += line_glue;
#ifdef VJDEBUG
fprintf(stderr,"line %d at %d contains %d (cf. %d), line_glue=%d, accum_skip=%d\n",
	pline, ps_line[pline], page_line[pline], pline * ps_baselineskip10 / 10, line_glue, accum_skip);
#endif
	    }

	    return;
	}

	if (!dojustify) return;

	pline = 0;
	while (page_line[end_line] == VGLUE) {
		for ( ; pline < end_line - 1; pline++) {
			if (page_line[pline] < 0 &&
			    penalty_line[pline] <= 0) break;
		}
		if (pline >= end_line - 1) break;
		vpads++;
		for (i = end_line; i > pline; i--) {
			page_line[i] = page_line[i-1];
			page_ins[i] = page_ins[i-1];
			penalty_line[i] = penalty_line[i-1];
		}
		pline++;
		pline++;
	}
	pline = 0;
	if (vpads > 2) while (page_line[end_line] == VGLUE) {
		for ( ; pline < end_line - 1; pline++) {
			if (page_line[pline] < 0 &&
			    penalty_line[pline] <= 0) break;
		}
		if (pline >= end_line - 1) break;
		for (i = end_line; i > pline; i--) {
			page_line[i] = page_line[i-1];
			page_ins[i] = page_ins[i-1];
			penalty_line[i] = penalty_line[i-1];
		}
		pline++;
		pline++;
	}
}

/*#define FFDEBUG*/
static int find_insertion(int curr_hbox, int pline, int how_high, int fn_printed) {
	int cond = FALSE;
	int ins_number, ins_type, other_insertions = 0;
	int total_lines, room, curr_fn_lines;
	int first_footnote = -1;

	room = how_high - pline;
	curr_fn_lines = 0;

	for (ins_number = 0; ins_number < MAXINSERT; ins_number++) {

		if (!insertion[ins_number].count) continue;
		ins_type = insertion[ins_number].atom;
		if (ins_type != FOOTNOTE) {
			other_insertions++;
			continue;
		}

		/* have we reached the peg? */
		if (prev_hbox(curr_hbox, insertion[ins_number].peg)) continue;

		total_lines = insertion[ins_number].total_lines;
		/* not first footnote? */
		if (fn_printed || curr_fn_lines) total_lines--;

#ifdef FFDEBUG
fprintf(stderr,"line %d: fn %d (1st goes at %d) fn_printed=%d\n", pline, ins_number,
		how_high - curr_fn_lines - total_lines, fn_printed);
#endif
		/* is there room for this footnote? */
		if (pline > how_high - curr_fn_lines - total_lines) break;

		/* yes, there's room */
		if (!curr_fn_lines) first_footnote = ins_number;
		curr_fn_lines += total_lines;
		if (fn_lines < curr_fn_lines) fn_lines = curr_fn_lines;

	}

	if (curr_fn_lines && pline == how_high - curr_fn_lines) return first_footnote;

	if (!other_insertions) return -1;

	for (ins_number = 0; ins_number < MAXINSERT; ins_number++) {

		if (!insertion[ins_number].count) continue;
		ins_type = insertion[ins_number].atom;
		if (ins_type == FOOTNOTE) continue;

		/* have we reached the peg? */
		if (ins_type != TOPINSERTION && prev_hbox(curr_hbox, insertion[ins_number].peg)) continue;

		if (ins_type == PAGEINSERTION) {
			if (pline == 0) {
				cond = TRUE;
				break;
			}
			continue;
		}
		room -= fn_lines;
		    /* is there room on this page? */
		total_lines = insertion[ins_number].total_lines;
		/* extra blank line at start of insertion */
		if (ins_type == MIDINSERTION && pline) total_lines++;
		/* extra blank line at end of insertion */
		if (ins_type == MIDINSERTION || ins_type == TOPINSERTION) total_lines++;

		if (  pline==0
		   || (total_lines <= room &&
		       (ins_type==MIDINSERTION || ins_type==VCENTER || ins_type==VBOX)
		      )
		   ) cond = TRUE;

		if (pline > 0 && ins_type==TOPINSERTION ) cond = FALSE;

		if (cond) break;
	}

	if (cond) return ins_number;
	else return -1;
}


/* #define TRACEPLAN */
static void plan_page() {
	static int main_hbox;
	int ins_number = -1, ins_count = 0;
	int i, pline, stuff = VGLUE, vskip = 0, vfillcnt = 0;
	int discarded_hboxes = 0, will_discard = VGLUE;
	int penalty = 0, min_penalty = 10000;
	int stretch = 0, underfull = VGLUE;
	int centering_space = 0, wh, ins_type = 0;
	int extra_skip = 0;
	int how_high = text_height;
	int needskips = TRUE, fn_printed = FALSE;

	/* don't put an extra hbox at the bottom of the page array */
	if (fflag) how_high--;

	/* fix this for insertions that take more than a page */
	wh = 0;

	i = first_hbox;
	main_hbox = i;

	fn_lines = 0;

	for (pline = 0; pline <= how_high; page_line[pline++] = stuff) {
#ifdef FFDEBUG
fprintf(stderr,"\t(top) hbox %d at line %d, i=%d (text_height=%d fn_lines=%d\n", stuff, pline, i, text_height, fn_lines);
#endif
		stuff = VGLUE;
		penalty_line[pline] = 0;
		/* ps_line[pline] = pline * ps_baselineskip10 / 10 + extra_skip; */
		ps_line[pline] = 0;
		/* finished an insertion? */
		if (wh && (i >= ins_count || pline >= how_high /*- 1*/)) {
#ifdef FFDEBUG
fprintf(stderr,"\tfinished ins %d at line %d\n", ins_number, pline);
#endif
			i = main_hbox;
			wh = 0; /* should check for more insertions */
			centering_space = 0;
			pending_insertions--;
			/* but what if I want to put back something?? */
			ins_count = insertion[ins_number].count = 0;
			ins_number = -1;
			if (!vskip && ins_type != FOOTNOTE) vskip++;
			if (ins_type==PAGEINSERTION) {
				vskip = how_high - pline;
				stuff = VFILL;
				/* ins_type = 0; */
				continue;
			}
			ins_type = 0;
			if (vskip) { vskip--; continue; }
		}
#if 0
		if (vskip > 0) {
			vskip--;
			continue;
		}
#endif
#ifdef DONOTIFY
			/* insertion notification? */
		if (!wh && hbox[0][i].atom == NOTIFY_ATOM) {
			#ifdef TRACEPLAN
			printf("\r\nNotify at line=%d\r\n",pline+1); getchr();
			#endif
			pending_insertions++;
			hbox[wh][i].atom = DEAD_ATOM;
			i++;
			if (i == MAXHBOX) i = LOWEST_HBOX;
			main_hbox = i;
		}
#endif

		if (!wh && pending_insertions > 0 && ins_type != PAGEINSERTION) {
			ins_number = find_insertion(i, pline, how_high, fn_printed);

			if (ins_number >= 0) {
			    /* must discard pending vskips to avoid passing over a
			     * line where a footnote should begin
			     */
			    vskip = 0;
			    ins_type = insertion[ins_number].atom;
			    ins_count = insertion[ins_number].count;
			    if (ins_type == VCENTER)
			      centering_space =
			  	(hsize_sp - SP(insertion[ins_number].maxlen)
				 - insertion[ins_number].minspaces)/2;
			    else centering_space = 0;
			    wh = insertion[ins_number].which;
			    main_hbox = i;
			    i = insertion[ins_number].first;
			    if (ins_type == FOOTNOTE) {
				    /* skip the horizontal rule after 1st footnote */
				    if (fn_printed) i++;
				    fn_printed = TRUE;
			    }
			    if (pline && (ins_type == MIDINSERTION || ins_type == TOPINSERTION))
					    continue;
#ifdef FFDEBUG
fprintf(stderr,"\tstart insertion %d of %d items at line %d\n", ins_number, ins_count, pline);
#endif
			    if (ins_type == PAGEINSERTION)
		 		hbox[wh][i].vglue = hbox[wh][i].vskip =
	  		         (how_high - insertion[ins_number].total_lines)/2;
			}
		}
		if (!wh && i == curr_hbox) {
			#ifdef TRACEPLAN
			if (pline == how_high - 2)
			{ printf("\r\nHit end, ins=%d, wh=%d, pend=%d\r\n",
			ins_number,wh,pending_insertions); getchr(); }
			#endif
			/*if (!pending_insertions) {*/
				if (fflag && !pending_insertions) stuff = ENDBOX;
				else if (vfillcnt) stuff = VGLUE;
				else stuff = VFILL;
			/*}*/
			continue;
		}

		if (!wh && psflag && extra_skip &&
				pline + extra_skip / ps_baselineskip >= how_high - fn_lines - 1)
			continue;
		if (vskip > 0) {
			vskip--;
			continue;
		}
		/* add requested vertical glue */
		if (needskips && hbox[wh][i].vskip > 0) {
			needskips = FALSE;
			vskip = hbox[wh][i].vskip;
			if (pline == 0 && !fflag) {
				if (hbox[wh][i].vglue) vskip = hbox[wh][i].vglue;
				else vskip = 0;
			}
			stretch += vskip;
			if (hbox[wh][i].vfill) {
				stuff = VFILL;
				vfillcnt++;
				if (!vskip) vskip = 1;
			}
			if (vskip > 0) {
				if (vskip + pline >= how_high - fn_lines && hbox[wh][i].penalty <= 0) {
					discarded_hboxes = 0;
					will_discard = VGLUE;
				}
				/* vskip--;
				 * continue; */
			}
		}
		if (vskip > 0) {
			vskip--;
			continue;
		}
		needskips = TRUE;

		if (hbox[wh][i].atom != HBOX_ATOM || !hbox[wh][i].ch) {
#ifdef DONOTIFY
			if (hbox[wh][i].atom == NOTIFY_ATOM) {
				if (pline > 0) pline--;
			}
			else {
				i++;
				if (i == MAXHBOX) i = LOWEST_HBOX;
			}
#else
			i++;
			if (i == MAXHBOX) i = LOWEST_HBOX;
#endif
		}
		else {
		/* add an hbox to the page */
		    stuff = i;

		    page_ins[pline] = wh;
				#ifdef TRACEPLAN
				printf("\r\nhbox=%d, wh=%d, line=%d\r\n",i,wh,pline); getchr();
				#endif
		    vskip = hbox[wh][i].vskip;

		    if (pline) {
		        extra_skip += hbox[wh][i].pskip;
		        ps_line[pline-1] += hbox[wh][i].pskip;
		    }
		    /* this may be a vcenter insertion */
		    hbox[wh][i].spaces += centering_space;

		    penalty = hbox[wh][i].penalty;
		    if (pline) {
			int fill_line = pline - 1;
			while (fill_line >= 0 && page_line[fill_line] == VGLUE) {
			    penalty_line[fill_line] = penalty;
			    fill_line--;
			}
		    }

		    if (!wh) {
			int line_factor = how_high - pline;

			if (fn_lines && pline < how_high) line_factor -= fn_lines;

			penalty +=
			  line_factor * (10*pagestretch/how_high) -
			  pagestretch * vskip / 20 - pagestretch * stretch / 10;
			penalty_line[pline] = penalty;

			if (penalty <= min_penalty) {
				min_penalty = penalty;
				discarded_hboxes = 1;
				will_discard = i;
				underfull = pline;
			}
			else discarded_hboxes++;
		    }
		    else {
			penalty =
			  penalty_line[pline] = 10000;
			discarded_hboxes = 0;
			will_discard = VGLUE;
			min_penalty = 10000;
		    }

		    vskip = 0;
		    if (hbox[wh][i].us) {
			/* if (xflag) vskip++; */
			if (!uflag && !oflag) vskip++;
		    }
		    if (baselineskip > 1)
			vskip += baselineskip - 1;

		    i++;
		    if (wh) {
			insertion[ins_number].first = i;
		    }
		    else {
			buffered_hboxes--;
			if (i == MAXHBOX) i = LOWEST_HBOX;
			main_hbox = i;
		    }
		}
#ifdef FFDEBUG
fprintf(stderr,"\thbox %d at line %d, i=%d (text_height=%d fn_lines=%d\n", stuff, pline, i, text_height, fn_lines);
#endif
	}

	if (fflag) {
		first_hbox = i;
		return;
	}

	/* this needs some thought */
	if (wh) {
			#ifdef TRACEPLAN
			printf("\r\nEnded page in insertion mode.\r\n");
			getchr();
			#endif

		fprintf(stderr,"\r\nEnded page in insertion mode.\r\n");
			getchr();
		if (ins_number >= 0) {
			ins_count = insertion[ins_number].count = 0;
			first_hbox = main_hbox;
			if (pending_insertions > 0) pending_insertions--;
		}
#if 0
		insertion[ins_number].first = i;
		i = main_hbox;
		first_hbox = main_hbox;
		wh = 0;
		pending_insertions--;
		/* but what if I want to put back something?? */
		ins_count = insertion[ins_number].count = 0;
		ins_number++;
		if (ins_number == MAXINSERT) ins_number = 0;
#endif
		vjustify(extra_skip);
		return;
	}

	if (fflag || will_discard == first_hbox ||
		 (i == curr_hbox && pending_insertions == 0)) {
			#ifdef TRACEPLAN
			printf("\r\nEnd or too much discard!\r\n"); getchr();
			#endif
			#ifdef FFDEBUG
			fprintf(stderr,"End or too much discard!\n");
			#endif
		/* if we hit the end, put back any hbox past the end of page */
		if (!fflag && i == curr_hbox && (stuff=page_line[text_height]) >= 0) {
			#ifdef FFDEBUG
			fprintf(stderr,"End: stuff=%d i=%d\n", stuff, i);
			#endif
			if (stuff != i) buffered_hboxes++;
			i = stuff;
		}
		first_hbox = i;
		vjustify(extra_skip);
		return;
	}

	if (will_discard >= 0 && (min_penalty < penalty || (penalty < 0 && discarded_hboxes == 1))) {
		first_hbox = will_discard;
		buffered_hboxes += discarded_hboxes;
			#ifdef TRACEPLAN
			printf("\r\nwill discard=%d, #disc.=%d\r\n",
	 		will_discard, discarded_hboxes);
			getchr();
			#endif
			#ifdef FFDEBUG
			fprintf(stderr,"will discard=%d, #disc.=%d\n",
	 		will_discard, discarded_hboxes);
			#endif
		while (underfull > 0 && page_line[underfull-1] < 0) underfull--;
		for (pline = underfull; pline <= text_height - fn_lines; pline++) {
			page_line[pline] = VGLUE;
			penalty_line[pline] = 0;
		}
	}
	else if ( (stuff=page_line[text_height]) >= 0) {
		first_hbox = stuff;
		buffered_hboxes++;
	}
	else first_hbox = i;

	#ifdef TRACEPLAN
	printf("\r\nEnded page plan with first = %d, i=%d, pending=%d\r\n",
	 first_hbox,i,pending_insertions);
	getchr();
	#endif

	vjustify(extra_skip);
}

static void copy_page() {
	int pline;
	if (fflag || !out_is_tty) return;
	for (pline = 0; pline < text_height; pline++) {
		last_page_line[pline] = page_line[pline];
		last_page_ins[pline] = page_ins[pline];
	}
}

static void do_space(int spaces) {
	int j;
	for (j = 0; j < spaces; j += COLWIDTH)
		schar(' ');
}

static void print_page(int last) {
	int pline, stuff, understuff, wh;

	if (psflag) pagesetup(curr_page);

	if (vflag) fprintf(stderr,"<%s", pageno_str);
	for (pline = 0; pline < text_height; pline++) {
		if (last) stuff = last_page_line[pline];
		else stuff = page_line[pline];
		if (stuff == ENDBOX) break;
		if (stuff >= 0) {
			if (last) wh = last_page_ins[pline];
			else wh = page_ins[pline];
			if (psflag) ps_outline(ps_line[pline], wh, stuff);
			else {
			    curr_col = 0;
			    do_space(hbox[wh][stuff].spaces + offset_sp);
			    outline(wh, stuff);
			}
			if (zflag && !psflag) printf("[%d]", penalty_line[pline]);
			if (last) understuff = last_page_line[pline + 1];
			else understuff = page_line[pline + 1];
			if (hbox[wh][stuff].us && pline + 1 < text_height &&
					!uflag && !oflag && understuff < 0) {
				NL();
				pline++;
			        curr_col = 0;
				do_space(hbox[wh][stuff].spaces + offset_sp);
				reputline(wh, stuff);
			}
		}
		else if (zflag && !psflag) printf("[%d]", penalty_line[pline]);
		NL();
	}
	if (psflag) pagefinish();
	if (vflag) fprintf(stderr,">");
}

static void do_top_margin() {
	int i, w;
	if (fflag || psflag) return;
	if (!out_is_tty && (!top_margin || outdev > 0) && curr_page > 1) putchar('\f');
	for (i = 0; i < top_margin - 3; i++) NL();
	if (top_margin >= 1) {
	    if (!nopagenumbers && pageattop) {
		fix_pageno();
		w = strlen(pageno_str);
		for (i = 0; i < offset + (hsize-w)/2; i++)
			putchar(' ');
		printf("-%s-", pageno_str);
	    }
	    NL();
	}
	if (top_margin > 1) NL();
	if (top_margin > 2) NL();
}
static void do_bot_margin() {
	int i, j, p, w;

	if (fflag || !bot_margin || psflag) return;

	if (!out_is_tty) {
		for (i = 0; i < bot_margin - 3; i++) NL();
		if (bot_margin >= 1) {
		    if (!nopagenumbers && !pageattop) {
			fix_pageno();
			w = strlen(pageno_str);
			for (i = 0; i < offset + (hsize-w)/2; i++)
				putchar(' ');
			printf("-%s-", pageno_str);
		    }
		    NL();
		}
	}

	if (bot_margin > 1) NL();
	if (bot_margin > 2) NL();
	if (out_is_tty) {
		if (!nopagenumbers) {
			p = 2 + strlen(pageno_str);
			j = offset + hsize/2 - p/2;
			for (i = 0; i < j; i++) putchar(' ');
			printf("-%s-", pageno_str);
			j = cols - j - p;
			for (i = 0; i < j - 8; i++) putchar(' ');
		}
		else for (i = 0; i < cols - 8; i++) putchar(' ');
		if (uflag) colorattr(c_prompt);
		printf("--MORE--\r");
		if (uflag) colorattr(c_roman);
		fflush(stdout);
		/* while ( getchar() != '\n') ; Use this if no TERMIOS? */
		i = getchr();
		if (uflag) termattr(CE);
		else NL();
		if (i == 'q' || i == ('C'& 0x1f)) {
			raw_mode(FALSE);
			if (kflag && !alwayskflag) printf("%c%%@", 27);
			exit(0);
		}
		else if (i == 'b' && !last_page_flag && curr_page > 2) {
			last_page_flag = TRUE;
		}
	}
	else NL();
	for (i = 0; i < bot_margin - 3; i++) NL();
}

static void shipout(int purge)
{

	if (buffered_hboxes < text_height + 2 && !purge) return;

	fix_pageno();
	if (psflag) job();

	while (curr_hbox != first_hbox || pending_insertions) {
		copy_page();
		plan_page();
		do_top_margin();
		print_page(FALSE);
		last_page_flag = FALSE;
		do_bot_margin();
		next_pageno();
		/* it ought to be possible to go `b' here on last page */
		while (last_page_flag) {
			prev_pageno();
			while (last_page_flag) {
				last_page_flag = FALSE;
				prev_pageno();
				do_top_margin();
				print_page(TRUE);
				do_bot_margin();
				next_pageno();
			}
			do_top_margin();
			print_page(FALSE);
			do_bot_margin();
			next_pageno();
		}
		if (purge && pending_insertions > 0) continue;
		if (!purge && buffered_hboxes < text_height + 2) break;
	}

	if (purge && psflag) endjob();
}

void shipline(unsigned char *oline, unsigned char *aline,
	int len, int tlen, int spaces, int vskip, int vglue, int pointskip,
	int underscoring, int penalty, int insnumber)
{
	static int last_insnumber = 0, last_peg = 1, total_pointskip = 0;
	int wh = 0, vfill = 0;
	int i, the_hbox, hbox_lines = 0;

	while (vskip >= 10000) {
		vskip -= 10000;
		vfill++;
	}
	if (vfill && !vskip) vskip++;
	if (vskip < vglue) vskip = vglue;
	if (psflag) {
		vskip += pointskip / ps_baselineskip;
		pointskip = pointskip % ps_baselineskip;
	}
	else pointskip = 0;

	if (fflag) insnumber = 0;

	if (insnumber != last_insnumber) {
		if (last_insnumber) { /* terminate the last one */
#ifdef DONOTIFY
			/* tell pager it's here */
			hbox[0][curr_hbox].atom = NOTIFY_ATOM;
			next_hbox();
#else
			pending_insertions++;
#endif
			if (total_pointskip)
			insertion[curr_ins].total_lines += 1 + total_pointskip / ps_baselineskip;
			curr_ins++;
			if (curr_ins == MAXINSERT) curr_ins = 0;
		}
		if (insnumber) { /* starting a new one */
			/* record the type of insertion */
			insertion[curr_ins].atom = insnumber & 0x0f;
			insertion[curr_ins].count = 0;
			insertion[curr_ins].first = 0;
			insertion[curr_ins].maxlen = 0;
			insertion[curr_ins].minspaces = 1000000;
			insertion[curr_ins].total_lines = 0;
			total_pointskip = 0;
			insertion[curr_ins].peg = last_peg;
			/* which should point to the right insert (??) */
			wh = next_ins_hboxes;
			insertion[curr_ins].which = wh;
			if (hbox[wh] == NULL) hbox[wh] =
			   (struct hboxtype*)malloc(
					 MAXPAGE * sizeof(struct hboxtype) );
			if (hbox[wh] == NULL) {
				windup();
				fprintf(stderr,"Couldn't get memory for insertion.");
				exit(1);
			}
			next_ins_hboxes++;
			if (next_ins_hboxes > MAXINSERT)
				next_ins_hboxes = 1;
		}
		else wh = 0;
		last_insnumber = insnumber;
	}
	else if (!insnumber) wh = 0;
	else wh = insertion[curr_ins].which;

	if (!wh) the_hbox = curr_hbox;
	else the_hbox = insertion[curr_ins].count;

	hbox[wh][the_hbox].atom = HBOX_ATOM;
		hbox_lines = 1;
	if (oline) hbox[wh][the_hbox].ch = keep_this(oline, len+1);
	else hbox[wh][the_hbox].ch = NULL;

/*
fprintf(stderr,"line %d spaces = %d (%d col), offset = %d\n", the_hbox, spaces, COL(spaces), offset_sp);
*/
	while (spaces + offset_sp < 0 && len > 0) {
		spaces += 500;
		len--;
		oline++;
		aline++;
	}
	if (spaces + offset_sp < 0) spaces = -offset_sp;
/*
fprintf(stderr,"and line %d spaces = %d (%d col), offset = %d\n", the_hbox, spaces, COL(spaces), offset_sp);
*/

	if (oline) {
		for (i = 0; i < len; i++) if (aline[i] != ROMAN) break;
		if (i >= len) hbox[wh][the_hbox].at = all_roman;
		else hbox[wh][the_hbox].at = keep_this(aline, len+1);
		hbox[wh][the_hbox].len = len;
		hbox[wh][the_hbox].tlen = tlen;
	}
	else {
		hbox[wh][the_hbox].at = NULL;
		hbox[wh][the_hbox].len = 0;
		hbox[wh][the_hbox].tlen = 0;
	}

	hbox[wh][the_hbox].us = underscoring;
		if (underscoring) {
			/*if (xflag) hbox_lines++;*/
			if (!uflag && !oflag) hbox_lines++;
		}
	hbox[wh][the_hbox].penalty = penalty;
	hbox[wh][the_hbox].vskip = vskip;
		hbox_lines += vskip;
	hbox[wh][the_hbox].pskip = pointskip;
	hbox[wh][the_hbox].vfill = vfill;
	hbox[wh][the_hbox].vglue = vglue;
	hbox[wh][the_hbox].spaces = spaces;
	hbox[wh][the_hbox].tb = tabid;

	if (wh == 0) {
		last_peg = the_hbox;
		total_pointskip = 0;
		next_hbox();
		buffered_hboxes++;
		shipout(NOPURGE);
	}
	else {
		the_hbox++;
		if (the_hbox >= MAXPAGE) {
			windup();
			fprintf(stderr,"Insertion too long.");
			exit(1);
		}
		insertion[curr_ins].count = the_hbox;
		insertion[curr_ins].total_lines += hbox_lines;
		total_pointskip += pointskip;
		if (len > insertion[curr_ins].maxlen)
			insertion[curr_ins].maxlen = len;
		if (spaces < insertion[curr_ins].minspaces)
			insertion[curr_ins].minspaces = spaces;
	}
}
#endif

/* terminal commands -- accessing routines */

/*
 *  following derived from:
 *  tcap v1.0 Enhanced terminal control program
 *  Author: Eric Lorenzo Lim
 *  Modified by: Becca Thomas and Rik Farrow
 *  (appeared in Nov. '87 Unix World)
 */

static char   *id[] = {
	"so", "se", "mb", "md", "mr", "mh", "me", "us", "ue",
	"as", "ae", "ac", "Sf", "ce",
	(char *) NULL,
};

static char   *area[50];
#ifndef HAVE_LIBNCURSES
static char    bp[2500];
#endif
static char    strbuf[1024];

/* get for later use terminal handling commands we need */
void terminit() {
	int     id_size = 0;
	char   *str;
#ifndef HAVE_LIBNCURSES
	char   *tmpptr;
#endif

#ifdef HAVE_LIBNCURSES
#define HAVE_ONE_OF_THEM
	extern char    *tgetstr();
#else
#ifdef HAVE_LIBTERMCAP
#define HAVE_ONE_OF_THEM
	extern char    *tgetstr();
#endif
#endif
	extern char    *getenv();

#ifndef HAVE_LIBNCURSES
	if ((tmpptr = getenv ("TERM")) == (char *) NULL) {
		(void) fprintf (stderr, "fm: TERM variable not set\n");
		exit (1);
	}
#endif

	str = strbuf;


#ifdef HAVE_ONE_OF_THEM
#ifdef HAVE_LIBNCURSES
	if (setupterm((char *)0, 1, (int *)0) == OK) {
#else
	if (tgetent (bp, tmpptr)) {
#endif
		while (id[id_size] != (char *) NULL) {
			area[id_size] = tgetstr (id[id_size], &str);
			if (area[id_size] == (char *) NULL)
				area[id_size] = "";
	/*else fprintf(stderr,"id %d [%s]\n", id_size, area[id_size]);*/
			id_size++;
		}
	}
	else {
		while (id[id_size] != (char *) NULL) {
				area[id_size] = "";
			id_size++;
		}
	}
#else
		while (id[id_size] != (char *) NULL) {
				area[id_size] = "";
			id_size++;
		}
#endif

#if HAVE_LIBNCURSES
	cols = columns;
	page_length = lines;
#else
	get_screensize(&page_length, &cols);
#endif
	cols_sp = SP(cols);
	if (!kflag && !aflag) {
	  int i, c;
	  i = 0;
	  while (area[AC][i]) {
		c = (unsigned char)area[AC][i];
		if (c < 128)
		  glyph[c] =  (unsigned char)area[AC][i+1];
		i += 2;
	  }
	}
}				/* terminit */

static void psglyph() {
/* "lceil" */		glyph['l'] = 'l'; glyph['L'] = 'L';
/* "rceil" */		glyph['k'] = 'k'; glyph['K'] = 'K';
/* "lfloor" */		glyph['m'] = 'm'; glyph['M'] = 'M';
/* "rfloor" */		glyph['j'] = 'j'; glyph['J'] = 'J';
/* "|" */		glyph['x'] = 'x'; glyph['X'] = 'X';
/* "top" */		glyph['w'] = 'w'; glyph['W'] = 'W';
/* "bot" */		glyph['v'] = 'v'; glyph['V'] = 'V';
/* "vdash" */		glyph['t'] = 't'; glyph['T'] = 'T';
/* "dashv" */		glyph['u'] = 'u'; glyph['U'] = 'U';
/* "--"	 */		glyph['q'] = 'q'; glyph['Q'] = 'Q';
/* \\"|" */		glyph['n'] = 'n'; glyph['N'] = 'N';
/* box */		glyph['h'] = 'h'; glyph['H'] = 'H';
}


static void open_getchr()
{
	tty = open("/dev/tty", O_RDONLY);
	if (tty < 0)
		tty = 2;
}

#ifndef STANDALONETREE

static int getchr()
{
	char c;
	int result;

	do
	{
		result = read(tty, &c, sizeof(char));
	} while (result != 1);

	return (c & 0377);
}

#include <termios.h>

static void raw_mode(int on)
{
	static int curr_on = FALSE;

	if (fflag) return;

	if (on == curr_on) return;

    {
	struct termios s;
	static struct termios save_term;
	static int saved_term = FALSE;

	if (on) {
		tcgetattr(tty, &s);

		if (!saved_term)
		{
			save_term = s;
			saved_term = TRUE;
		}
		cfmakeraw(&s);

	}
       	else {
		s = save_term;
	}
	tcsetattr(tty, TCSADRAIN, &s);
    }
	curr_on = on;
}

void calculate_offset() {
	/* Default lmargin is 6000 = SP(12). */
	if (psflag && !out_is_tty) {
		lmargin = 2 * ps_pitch_sp / 3;
	}
	offset_sp = lmargin + hoffset;
#ifdef DEBUGMARGINS
fprintf(stderr,"offset_sp=%.2f <- lmargin %.2f + hoffset %.2f\n",
		PT(offset_sp)/72.0, PT(lmargin)/72.0, PT(hoffset)/72.0);

fprintf(stderr,"\t2*offset_sp=%.2f + hsize_sp=%.2f > cols_sp=%.2f\n",
		PT(offset_sp)/36.0, PT(hsize_sp)/72.0, PT(cols_sp)/72.0);
#endif
	if (2*offset_sp + hsize_sp > cols_sp)
		  offset_sp = (cols_sp - hsize_sp) / 2;
#ifdef DEBUGMARGINS
fprintf(stderr,"new offset_sp=%.2f <- (cols_sp %.2f - hsize_sp %.2f)/2\n",
		PT(offset_sp)/72.0, PT(cols_sp)/72.0, PT(hsize_sp)/72.0);
#endif
	if (offset_sp < 0) offset_sp = 0;
	offset = COL(offset_sp);
}
void calculate_ps_page_length() {
	if (!psflag) return;
	/* rounding down ?? */
	page_length = (int)(792.0 / ps_baselineskip);
	top_margin = 72 / ps_baselineskip;
	if (vsize < 0) {
		bot_margin = top_margin;
		vsize = page_length - top_margin - bot_margin;
	}
	else bot_margin = (page_length - vsize * 72) / ps_baselineskip;
	if (bot_margin < 0) bot_margin = 0;
	/*top_margin = bot_margin = 72 / ps_baselineskip;*/
	/*vsize = page_length - top_margin - bot_margin;*/
	text_height = vsize;
}
#endif

void ttyinit() {
	out_is_tty = isatty(1);
	in_is_tty = isatty(0);
	open_getchr();
	if (psflag || kflag) psglyph();
}

#ifdef HAVE_TERMIOS_H
#define HAVE_SIZECHANGE
#endif
#ifdef HAVE_SYS_IOCTL_H
#ifdef GWINSZ_IN_SYS_IOCTL
# include <sys/ioctl.h>
#endif
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
extern int errno;
#else
#undef HAVE_SIZECHANGE
#endif

/* Obtain lines/columns values from the environment and/or terminfo entry */
static void get_screensize(int *linep, int *colp) {
	char	*rows, *cols;

	*linep = *colp = 0;

#ifdef HAVE_SIZECHANGE
	/* first try asking the OS */
	if (*linep <= 0 || *colp <= 0) {
	    struct winsize size;

	    errno = 0;
	    do {
		if (ioctl(tty, TIOCGWINSZ, &size) < 0
			&& errno != EINTR)
		    goto failure;
	    } while (errno == EINTR);

	    *linep = (int)size.ws_row;
	    *colp  = (int)size.ws_col;
	failure:;
	}
#endif /* HAVE_SIZECHANGE */

	/* if that didn't work, try environment variables */
	if (*linep <= 0 || *colp <= 0) {
		extern char    *getenv();
		rows = getenv("LINES");
		if (rows != 0)
			*linep = atoi(rows);
		cols = getenv("COLUMNS");
		if (cols != 0)
			*colp = atoi(cols);
	}

	/* the ultimate fallback, assume fixed 24x80 size */
	if (*linep <= 0 || *colp <= 0) {
		*linep = 24;
		*colp  = 80;
	}
}


void devinit() {
	int i;
	static int initialized = FALSE;

	if (!initialized) {

#ifdef USESIGNALS
		if (signal(SIGINT, windup) == SIG_IGN)
			signal(SIGINT, SIG_IGN);
		if (signal(SIGTERM, windup) == SIG_IGN)
			signal(SIGTERM, SIG_IGN);
#endif

#ifndef STANDALONETREE
		buf = bufp = (unsigned char *)malloc(BUFFERSIZE);
		if (buf == NULL) exit(1);
		hbox[0] = (struct hboxtype*)malloc(
				 MAXHBOX * sizeof(struct hboxtype) );
		if (hbox[0] == NULL) exit(1);

		if (out_is_tty) raw_mode(1);
		for (i = 0; i < MAXLINE; i++)
			all_roman[i] = ROMAN;
#endif
		if (/*out_is_tty &&*/!psflag && !fflag && !uflag) {
			get_screensize(&page_length, &cols);
		}
		cols_sp = SP(cols);
		if (out_is_tty && kflag && !alwayskflag) printf("%c%%G", 27);
	}

#ifndef STANDALONETREE
	if (fflag) top_margin = bot_margin = 0;
	else if (!out_is_tty) {
		/* real paper? */
		if (psflag || outdev) {
			if (psflag) setfontnames();
			calculate_ps_page_length();
			if (!initialized) {
			    /* cols_sp = (int)( (8.5 - 0.33) * ps_pitch_sp + 0.5 );*/
			    /* cols_sp = ( (850 - 33) * ps_pitch_sp + 50 )/100; */
			    cols_sp =  850 * ps_pitch_sp /100;
#ifdef DEBUGMARGINS
fprintf(stderr," >> cols_sp=%.2f\n", PT(cols_sp)/72.0);
#endif
			    cols = COL(cols_sp);
			    /* hsize_sp = (int)(6.5 * ps_pitch_sp + 0.5); */
			    hsize_sp = (65 * ps_pitch_sp + 5)/10;
			    hsize = COL(hsize_sp);
			}
		}
		/* in case |less -r */
		else if (vsize < 0) {
			if (!nopagenumbers) page_length--;
			vsize = page_length - 1;
			if (!nopagenumbers) top_margin = 1;
			else top_margin = 0;
			bot_margin = 0;
		}
		else {
			if (vsize > page_length) vsize = page_length;
			if (vsize > 40) top_margin = bot_margin = 4;
			else if (vsize > 25) top_margin = bot_margin = 2;
			else { top_margin = 1; bot_margin = 0; }
		}
	}
	else {
		top_margin = 0;
		bot_margin = 1;
	}

	if (vsize < 0 || out_is_tty)
		 vsize = page_length - top_margin - bot_margin;

	text_height = vsize;

	calculate_offset();
#endif
	initialized = TRUE;
}
static int outch (int c)
{
	return putchar(c);
}

/* put out a terminal handing command */
/* thanks to Dave Yearke, Sigma Systems Technology, Inc.
 *   for the SYSVR3 change below
 */
static void termattr (int i)
{
#ifdef SYSVR3
	(void) putp (area[i]);
#else /* SYSVR3 */
#ifdef HAVE_ONE_OF_THEM
	if (area[i][0])
	(void) tputs (area[i], 1, outch);
	else
#endif
	switch(i) {
	case SO: printf("%c[7m", 27); break; /* smso */
	case SE: printf("%c[27m", 27); break; /* rmso */
	case MB: printf("%c[5m", 27); break; /* blink */
	case MD: printf("%c[1m", 27); break; /* bold */
	case MR: printf("%c[7m", 27); break; /* rev */
	case MH: printf("%c[2m", 27); break; /* dim */
	case ME: printf("%c[0;10m", 27); break; /* sgr0 =\E[m\017 for xterm */
	case US: printf("%c[4m", 27); break; /* smul */
	case UE: printf("%c[24m", 27); break; /* rmul */
	default: break;
	}
#endif /* SYSVR3 */
}				/* termattr */

int
	c_bold = C_RED,
	c_bitalic = C_MAGENTA,
	c_italic = C_CYAN,
	c_math = C_BLUE | M_REVERSE,
	c_ttype = C_GREEN | M_REVERSE,
	c_sans = C_BLUE,
	c_musix = C_RED | M_REVERSE,
	c_slanty = C_YELLOW,
	c_super = C_CYAN | M_REVERSE,
	c_ipa = C_GREEN,
	c_sub = C_YELLOW | M_REVERSE,
	c_prompt = C_WHITE | M_REVERSE,
	c_roman = C_WHITE;

static int last_color = 0;
static void colorattr(int i)
{
    int color = i & 0x07;
    static int lmode = 0;
    int mode;

    if (color == 0) color = C_WHITE;

    mode = (i >> 4) & 0x0f;

    if (mode != lmode) {
	if (lmode == US) termattr(UE);
	else if (lmode) { termattr(ME); last_color = C_WHITE; }
	lmode = 0;
    }

    if (color != last_color) {
        last_color = color;

        if (color == C_WHITE) {
	    /* printf("%c[m", 27); */
	    termattr(ME);
	    if (lmode != US) lmode = 0;
        }
    #ifdef HAVE_LIBNCURSES
        else if (area[Sf][0])
    	    putp(tparm(area[Sf], color));
    #endif
        else
	    printf("%c[3%cm", 27, '0' + color);
    }

    if (mode != lmode) {
	if (mode) termattr(mode);
	lmode = mode;
    }
}

/* check for change of "font" and send command to terminal
 * when there is a change */
#ifndef STANDALONETREE
static
#endif
void newattr(unsigned char attr)
{
	static unsigned char l = ROMAN;
	unsigned char a, b;

	a = attr;
	if (a >= 128) {
		a -= 128;
		if (!alt && !kflag) {
			termattr(AS);
			alt = 1;
		}
	}
	else {
		if (alt && !kflag) {
			termattr(AE);
			alt = 0;
		}
	}
	if (l >= 128) b = l - 128;
	else b = l;

	if (nflag) {
	    if (b != a) {

/*
		if (b == BOLD || b == TTYPE || b == MATH || b == SLANTY)
*/
			termattr (ME);
			last_color = C_WHITE;
/*
		else if (b == ITALIC)
			termattr (UE);
*/
		switch (a) {
			case  SUPER:
			case IPASUPER:
			case BITALIC:
			case   BOLD:	termattr (MD); break;
			case    IPA:
			case    SUB:
			case IPASUB:
			case ITALIC:	termattr (MH); break;
			case   SANS:
			case  MUSIX:
			case SLANTY:	termattr (MH); break;
			case   MATH:
			case  TTYPE:
					termattr (MR); break;
			default: break;
		}
	    }
	}

	else { /* nflag */
	    if (b != a) {
		switch (a) {
			case   BOLD: colorattr(c_bold); break;
			case ITALIC: colorattr(c_italic); break;
			case   MATH: colorattr(c_math); break;
			case  TTYPE: colorattr(c_ttype); break;
			case SLANTY: colorattr(c_slanty); break;
			case IPASUPER:
			case  SUPER: colorattr(c_super); break;
			case    IPA: colorattr(c_ipa); break;
			case IPASUB:
			case    SUB: colorattr(c_sub); break;
	    		case ZD:
			case BITALIC: colorattr(c_bitalic); break;
			case   SANS: colorattr(c_sans); break;
			case  MUSIX: colorattr(c_musix); break;
	    		case BOX:
			case  ROMAN:
			default: colorattr(c_roman); break;
		}
	    }

	} /* else nflag */

	l = attr;
}				/* newattr */
